/*
 * tree.cc
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>

#if STHREADS && !MTCAMP
#include <st.h>
#endif

// 1 if we build a fresh tree at every iteration
#define BUILD_TREE_EVERY_ITERATION 1
// 1 if we build a tree in parallel
#define BUILD_TREE_PARALLEL 1
// 1 if we gcollect at every iteration
#define GCOLLECT_EVERY_ITERATOIN 1

#ifdef WITH_SGC
#include <sgc_cpp.h>
struct maybe_gc : gc {};
#else
struct maybe_gc {};
#endif

template<class T>
struct tree : maybe_gc
{
  tree (T v) { leaf = 1; val = v; }
  tree (tree<T> * l, tree<T> * r) { leaf = 0; left = l; right = r; }
  int leaf;
  struct tree<T> * left, * right;
  T val;
};

/* serial make tree */
template<class T>
tree<T> * s_make_tree (int n, T v)
{
  if (n == 0) {
    tree<T> * t = new tree<T> (v);
    return t;
  } else {
    tree<T> * l = s_make_tree (n - 1, v);
    tree<T> * r = s_make_tree (n - 1, v);
    tree<T> * t = new tree<T> (l, r);
    return t;
  }
}

/* parallel make tree */
template<class T>
#if MTCAMP || !STHREADS
tree<T> *
#else
void 
#endif
p_make_tree (int n, T v
#if STHREADS && !MTCAMP
	     , tree<T> ** res, st_join_counter_t * c
#endif
		  )
{
  if (n == 0) {
    tree<T> * t = new tree<T> (v);
#if MTCAMP
    return t;
#elif STHREADS
    * res = t;
    st_join_counter_finish(c);
#else  /* SEQUENTIAL */
    return t;
#endif
  } else {
    tree<T> * l;
    tree<T> * r;
#if MTCAMP
    mtc_sync {
      ST_POLLING();
      mtc_fork l = p_make_tree(n - 1, v);
      r = p_make_tree(n - 1, v);
      ---;
      return new tree<T>(l, r);
    }
#elif STHREADS
    st_join_counter_t cc[1];
    st_join_counter_init(cc, 2);
    ST_POLLING();
    ST_THREAD_CREATE(p_make_tree (n - 1, v, &l, cc));
    p_make_tree (n - 1, v, &r, cc);
    st_join_counter_wait(cc);
#else  /* SEQUENTIAL */
    l = p_make_tree(n - 1, v);
    r = p_make_tree(n - 1, v);
#endif
    tree<T> * t = new tree<T> (l, r);

    ST_POLLING();
#if MTCAMP
    return t;
#elif STHREADS
    * res = t;
    st_join_counter_finish(c);
#else  /* SEQUENTIAL */
    return t;
#endif
  }
}

template<class T>
tree<T> * make_tree (int n, T v)
{
#if BUILD_TREE_PARALLEL
#if MTCAMP
    tree<T> * res = p_make_tree(n, v);
    ST_POLLING();
    return res;
#elif STHREADS
    tree<T> * res;
    st_join_counter_t c[1];
    st_join_counter_init(c, 1);
    p_make_tree(n, v, &res, c);
    st_join_counter_wait(c);
    ST_POLLING();
    return res;
#else  /* SEQUENTIAL */
    tree<T> * res = p_make_tree(n, v);
    return res;
#endif

#else  /* BUILD_TREE_PARALLEL */
    return s_make_tree(n, v);
#endif  /* BUILD_TREE_PARALLEL */
}

inline int combine(int x, int y) { return x + y; }
#define TTYPE int
#define LEAF_VAL 1

int block_frequency = 0;

template<class T>
#if MTCAMP || !STHREADS
T
#else
void 
#endif
p_sum (tree<T> * t, int code
#if STHREADS && !MTCAMP
       , T * rv, st_join_counter_t * ack
#endif
       )
{
  if (t->leaf) {
#ifdef BLOCK_TEST
    if (block_frequency) {
      if (code % block_frequency == 0) st_temporal_suspend();
    }
#endif

#if MTCAMP
    return t->val;
#elif STHREADS
    *rv = t->val;
    st_join_counter_finish(ack);
#else  /* SEQUENTIAL */
    return t->val;
#endif

  } else {
#if MTCAMP
    mtc_sync {
	T l, r;
	ST_POLLING();
	mtc_fork l = p_sum(t->left, 2 * code);
	r = p_sum(t->right, 2 * code + 1);
	---;
	ST_POLLING();
	return combine(l, r);
    }
#elif STHREADS
    int crv[2];
    st_join_counter_t cack[1];
    
    st_join_counter_init(cack, 2);
    ST_THREAD_CREATE(p_sum (t->left, 2 * code, crv, cack));
    p_sum(t->right, 2 * code + 1, crv + 1, cack);
    st_join_counter_wait(cack);
    *rv = combine (crv[0], crv[1]);
    ST_POLLING();
    st_join_counter_finish(ack);
#else
    T l = p_sum(t->left, 2 * code);
    T r = p_sum(t->right, 2 * code + 1);
    return combine(l, r);
#endif
  }
}

template<class T>
T sum (tree<T> * t)
{
#if MTCAMP
    T r = p_sum(t, 1);
#elif STHREADS
    T r;
    st_join_counter_t ack[1];
    st_join_counter_init(ack, 1);
    p_sum(t, 1, &r, ack);
    st_join_counter_wait(ack);
#else
    T r = p_sum(t, 1);
#endif
    ST_POLLING();
    return r;
}

/* build tree and scan it after extending a stack N times, so that
   junks are not left near the bottom of the stack */
template<class T>
void one_iteration(tree<T> * t, int d, int iter, int n, int * u)
{
    if (n == 0) {
	printf("stack size before %d th iteration = %d\n", 
	       iter, st_stack_used_bytes());
	long t0 = st_current_time_us();
#if BUILD_TREE_EVERY_ITERATION
	st_assert(t == 0);
	t = make_tree(d, LEAF_VAL);
#endif
	TTYPE r = sum(t);
	long t1 = st_current_time_us();
	printf ("%d %u -- %u = %d usec\n", r, t0, t1, t1 - t0);
    } else {
	int unused[100];
	one_iteration(t, d, iter, n - 1, unused);
    }
}

int st_main(int argc, char ** argv, char ** env)
{
#ifdef BLOCK_TEST
  if (argc < 4) {
    fprintf(stderr, "usage: %s depth n_iters block_freq\n", argv[0]);
    st_wg_exit((void *)1);
  }
#else
  if (argc < 3) {
    fprintf(stderr, "usage: %s depth n_iters\n", argv[0]);
    st_wg_exit((void *)1);
  }
#endif

  int d = atoi (argv[1]);	// depth of the tree
  int n = atoi (argv[2]);	// number of iterations
  int i;
#if BUILD_TREE_EVERY_ITERATION
  tree<TTYPE> * t = 0;
#else
  tree<TTYPE> * t = make_tree(d, LEAF_VAL);
#endif
  
#ifdef BLOCK_TEST
  block_frequency = atoi(argv[3]);
#endif
  for (i = 0; i < n; i++) {
      one_iteration(t, d, i, 10, 0);
#if GCOLLECT_EVERY_ITERATOIN
      fprintf(st_errout, "sleep a second\n");
      st_sleep_os_thread_ms(1000);
      /* at this point, there is only one busy worker (i.e., this worker),
	 and other workers have probably cleared stack. */
      st_remove_exported_frames_user();
      fprintf(st_errout, 
	      "worker %d : clear stack before collect (stack size = %d bytes)\n", 
	      tls(worker_id), st_stack_used_bytes());
      SGC_CLEAR_STACK();
      fprintf(st_errout, "worker %d : initiate collect\n", tls(worker_id));
      GC_gcollect();
      fprintf(st_errout, "worker %d : collect done\n", tls(worker_id));
#endif
  }
  return 0;
}

