/* 
 * racedtct.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include <racedtct.h>
#define PRIVATE static
#define GLOBAL 

/* by default, dont stop after the first race has been detected */
#ifndef STOP_AFTER_FIRST_RACE
#define STOP_AFTER_FIRST_RACE 0
#endif

/* 
 * Disjoint-set data structure
 */

/* a tag that distinguishes S bag and P bag */
typedef enum an_sp_bag_kind { 
  sbk_S, 
  sbk_P 
} an_sp_bag_kind;

#if defined RD_DEBUG
typedef struct children_set* children_set_t;
#endif /* RD_DEBUG */

/* an element of a disjoint set */
typedef struct disj_set
{
  int id;
  int rank;
  struct disj_set * link;	/* link to the parent */
#if defined RD_DEBUG
  children_set_t children;	/* link to the children */
#endif /* RD_DEBUG */
  an_sp_bag_kind SP;		/* 0 for S, 1 for P */
#if MEM_TRACE_PUT_ID
  source_id_t source_id;	/* source identifier of this thread */
#endif /* MEM_TRACE_PUT_ID */
  char * file;			/* where it is created */
  int line;
} * disj_set_t;

#if defined RD_DEBUG
struct children_set
{
  disj_set_t a_child;
  children_set_t other_children;
};
#endif /* RD_DEBUG */

/* 
 * active block stack
 */

/* a tag that distinguishes fork statement and sync block */
typedef enum a_mtc_block_kind {
  mbk_fork,			/* fork statement */
  mbk_sync			/* sync block */
} a_mtc_block_kind;

/* 
 * a record that describes fork or sync block.
 */
typedef struct mtc_block
{
  a_mtc_block_kind kind;	/* fork or sync */
  disj_set_t id_set;		/* singleton set that represents this block */
  disj_set_t S;			/* S bag */
  disj_set_t P;			/* P bag */
#if MEM_TRACE_PUT_ID
  source_id_t source_id;
#endif
} * mtc_block_t;

/* 
 * Data structures for recording reads and writes 
 */

typedef enum an_access_kind { 
  ak_read,
  ak_write,
  ak_last
} an_access_kind;

/* 
 * record associated with each location
 */
typedef struct TR_access_record
{
  disj_set_t th;		/* singleton set = thread id */
#if MEM_TRACE_PUT_ID
  source_id_t source_id;	/* source identifier of this access */
#endif
  char * file;
  int line;
} * TR_access_record_t;

/* record associated with an address */

typedef struct TR_trace_record
{
  struct TR_access_record rec[ak_last];	/* reader and writer */
} * TR_trace_record_t;

/* 
 * Page tables and TLB-like structures for mapping addresses 
 */

/* an address (assumed to be 32 bits for now) is broken down 
   into three parts (lowest 12 bits, middle 10 bits, and highest 
   10 bits */
#define TR_LOG_PAGE_SIZE 12
#define TR_PAGE_SIZE (1 << TR_LOG_PAGE_SIZE)
#define TR_PRIMARY_IDX_WIDTH 10
#define TR_PRIMARY_IDX_MASK ((1 << TR_PRIMARY_IDX_WIDTH) - 1)
#define TR_SECONDARY_IDX_WIDTH 10
#define TR_SECONDARY_IDX_MASK ((1 << TR_SECONDARY_IDX_WIDTH) - 1)

/* granularity (currently 4 bytes) */
#define TR_LOG_TRACE_GRANULARITY 2
#define TR_TRACE_GRANULARITY (1 << TR_LOG_TRACE_GRANULARITY)

/* mapping: address -> key = highest 22 bits */
#define TR_TLB_KEY(a) \
((((long)a) >> TR_LOG_PAGE_SIZE) & ((1 << TR_SECONDARY_IDX_WIDTH) - 1))
#define TR_TLB_OFFS(a) (((long)a) & (TR_PAGE_SIZE - 1))

/* mapping: address -> primary idx = highest 10 bits */
#define TR_PRIMARY_IDX(a) \
((((long)a) >> (TR_LOG_PAGE_SIZE + TR_SECONDARY_IDX_WIDTH)) \
 & TR_PRIMARY_IDX_MASK)

/* mapping: address -> secondary idx = middle 10 bits */
#define TR_SECONDARY_IDX(a) \
((((long)a) >> TR_LOG_PAGE_SIZE) & TR_SECONDARY_IDX_MASK)

typedef struct TR_secondary_idx
{
  void * idx[1 << TR_SECONDARY_IDX_WIDTH];
} * TR_secondary_idx_t;

typedef struct TR_tlb_entry 
{
  void * tag;
  void * val;
} * TR_tlb_entry_t;

typedef struct _TR_array
{
  struct TR_tlb_entry __tlb[1 << TR_SECONDARY_IDX_WIDTH];
  TR_secondary_idx_t __ptable[1 << TR_PRIMARY_IDX_WIDTH];
  /* the current thread number */
  unsigned __current_thread;
  /* size of a single trace record rounded up to a power of two */
  int __log_trace_record_size;
  int __trace_record_size;
  /* memory discarded for each page trace */
  int __trace_page_overhead;
  /* memory needed to record logs for a single page */
  int __trace_page_size;
  /* the amount of left shifts needed to scale offset within a page
     to the log page */
} * _TR_array_t;


/* keep track of races found */
typedef struct source_info_aux
{
  source_id_t id;
  char * file;
  int line;
} source_info_aux, * source_info_aux_t;

typedef struct source_info
{
  source_info_aux th;
  source_info_aux ac;
} * source_info_t;

typedef struct race_record
{
  struct race_record * next;
  struct source_info si[2];
} * race_record_t;

#define RACE_RECORD_HASH_IGNORE 6
#define RACE_RECORD_HASH_WIDTH 8
#define RACE_RECORD_HASH_SIZE (1 << RACE_RECORD_HASH_WIDTH)
#define RACE_RECORD_HASH_MASK (RACE_RECORD_HASH_SIZE - 1)

/* 
 * all globals are accessed through rd_globals
 */

typedef struct rd_globals
{
  int __inited;
  int __dont_race_detect;
  mtc_block_t __active_block_stack_p;	/* the top of the active_block_stack */
  mtc_block_t __active_block_stack_base; /* base of the active_block_stack */
  mtc_block_t __active_block_stack_lim; /* limit of the active_block_stack */
  disj_set_t __disj_set_space_p; /* allocation pointer for disjoint set */
  disj_set_t __disj_set_space_base; /* base for disjoint set allocation */
  disj_set_t __disj_set_space_lim; /* base for disjoint set allocation */
#if defined RD_DEBUG
  children_set_t __child_link_space_p;
  children_set_t __child_link_space_base;
  children_set_t __child_link_space_lim;
#endif /* RD_DEBUG */
  int __unique_fork_id;		/* forked procedure ID */
  race_record_t * __races_found; /* a pointer to a large array of 
				    races found (initially zero) */
  struct _TR_array __TR;	/*  */
} * rd_globals_t;

PRIVATE struct rd_globals _rd_globals[1] = { 
  {
    /*__inited=*/ 0, 
    /*__dont_race_detect=*/0 
  }
};
PRIVATE rd_globals_t __rd_globals = _rd_globals;

/* macros that access global structures */
#define rd_globals_inited (__rd_globals->__inited)
#define rd_dont_race_detect (__rd_globals->__dont_race_detect)

#define active_block_stack_p (__rd_globals->__active_block_stack_p)
#define active_block_stack_base (__rd_globals->__active_block_stack_base)
#define active_block_stack_lim (__rd_globals->__active_block_stack_lim)
#define disj_set_space_p (__rd_globals->__disj_set_space_p)
#define disj_set_space_base (__rd_globals->__disj_set_space_base)
#define disj_set_space_lim (__rd_globals->__disj_set_space_lim)
#if defined RD_DEBUG
#define child_link_space_p (__rd_globals->__child_link_space_p)
#define child_link_space_base (__rd_globals->__child_link_space_base)
#define child_link_space_lim (__rd_globals->__child_link_space_lim)
#endif /* RD_DEBUG */
#define unique_fork_id (__rd_globals->__unique_fork_id)
#define races_found (__rd_globals->__races_found)

#define TR_TLB (__rd_globals->__TR.__tlb)
#define TR_ptable (_rd_globals->__TR.__ptable)
#define TR_trace_record_size (_rd_globals->__TR.__trace_record_size)
#define TR_log_trace_record_size (_rd_globals->__TR.__log_trace_record_size)
#define TR_trace_page_overhead (_rd_globals->__TR.__trace_page_overhead)
#define TR_trace_page_size (_rd_globals->__TR.__trace_page_size)

/* Fatal error occurs */

PRIVATE void fatal_error_exit()
{
  exit(1);
}

/* general memory allocation */

PRIVATE void * mtc_general_alloc_1(size_t size, char * file, int line)
{
  void * p = (void *)malloc(size);
  if (p) return p;
  else {
    fprintf(stderr, "mtc_general_alloc failed at %s:%d\n", file, line);
    fatal_error_exit();
    return 0;			/* never reach */
  }
}
#define mtc_general_alloc(size) mtc_general_alloc_1(size, __FILE__, __LINE__)

PRIVATE int int_log(int x)
{
  if (x == 0) return -1;
  else {
    int i = 0;
    int s = 1;
    while (s < x) {
      s += s;
      i++;
    }
    if (x <= (1 << i) && x > (1 << (i - 1))) return i;
    else {
      fprintf(stderr, "bomb: log_int (%d) is about to return %d\n", 
	      x, i);
      fatal_error_exit();
      return 0;			/* never reach */
    }
  }
}

/* initialization */

#define INITIAL_ACTIVE_BLOCK_STACK_DEPTH (1 << 10)
#define INITIAL_SET_SPACE_SIZE (1 << 15)
#if defined RD_DEBUG
#define INITIAL_LINK_SPACE_SIZE (1 << 15)
#endif /* RD_DEBUG */

PRIVATE void init_TR_globals()
{
  int l = int_log (sizeof (struct TR_trace_record));
  int i;
  for (i = 0; i < (1 << TR_PRIMARY_IDX_WIDTH); i++)
    TR_ptable[i] = 0;
  TR_log_trace_record_size = l;
  TR_trace_record_size = (1 << l);
  TR_trace_page_overhead = 0;
  /* xxx: what if TR_GRANULARITY > TR_PAGE_SIZE? */
  TR_trace_page_size 
    = (TR_PAGE_SIZE << (l - TR_LOG_TRACE_GRANULARITY)) 
      + TR_trace_page_overhead;
#ifdef RD_DBG
  printf ("Trace granularity = %d bytes\n", TR_TRACE_GRANULARITY);
  printf ("Page size = %d bytes\n", TR_PAGE_SIZE);
  printf ("Record size = %d bytes\n", TR_trace_record_size);
  printf ("Record page size = %d bytes\n", TR_trace_page_size);
  printf ("Approximately %d%% space overhead expected\n",
	    (int)((double)TR_trace_page_size / (double)TR_PAGE_SIZE * 100.0));
#endif
}

#define INITIAL_THREAD_ID 0
GLOBAL void __rd_raw_fork_begin();

PRIVATE void init_rd_globals()
{
  /* allocate stack space for active blocks */
  int absd = INITIAL_ACTIVE_BLOCK_STACK_DEPTH;
  mtc_block_t active_block_stack
    = (mtc_block_t)mtc_general_alloc(sizeof(struct mtc_block) * absd);
  int sss = INITIAL_SET_SPACE_SIZE;
  disj_set_t set_space
    = (disj_set_t)mtc_general_alloc(sizeof(struct disj_set) * sss);
#if defined RD_DEBUG
  int lss = INITIAL_LINK_SPACE_SIZE;
  children_set_t child_link_space
    = (children_set_t)mtc_general_alloc(sizeof(struct children_set) * lss);
#endif /* RD_DEBUG */

  active_block_stack_p = active_block_stack;
  active_block_stack_base = active_block_stack;
  active_block_stack_lim = active_block_stack + absd;
  disj_set_space_p = set_space;
  disj_set_space_base = set_space;
  disj_set_space_lim = set_space + sss;
#if defined RD_DEBUG
  child_link_space_p = child_link_space;
  child_link_space_base = child_link_space;
  child_link_space_lim = child_link_space + lss;
#endif /* RD_DEBUG */
  unique_fork_id = INITIAL_THREAD_ID - 1;
  races_found = 0;

  init_TR_globals();

  rd_globals_inited = 1;

  /* make initial (main) thread */
  
#if MEM_TRACE_PUT_ID
  MEM_TRACE_RAW_FORK_BEGIN(0);
#else  /* MEM_TRACE_PUT_ID */
  MEM_TRACE_RAW_FORK_BEGIN();
#endif /* MEM_TRACE_PUT_ID */
}

GLOBAL void __rd_race_detect_on()
{
  rd_dont_race_detect = 0;
}

GLOBAL void __rd_race_detect_off()
{
  rd_dont_race_detect = 1;
}

/* Disjoint data structure implementation */

#define RD_EMPTY_SET ((disj_set_t)0)
#define RD_IS_EMPTY_SET(s) ((s) == RD_EMPTY_SET)
#define RD_IS_PBAG(s) ((s)->SP == sbk_P)
#define RD_IS_SBAG(s) ((s)->SP == sbk_S)

/* currently GC is not implemented */
PRIVATE void disj_set_space_gc()
{
  fprintf (stderr, "disjoing set space exhausted\n");
  fatal_error_exit();
}

PRIVATE disj_set_t alloc_disj_set()
{
  disj_set_t p;
  disj_set_t l;
  if (rd_globals_inited == 0) init_rd_globals();
  p = disj_set_space_p;
  l = disj_set_space_lim;
  if (p < l) {
    disj_set_space_p = p + 1;
    return p;
  } else {
    disj_set_space_gc();
    fprintf (stderr, "disj_set_space_gc\n"); 
    fatal_error_exit();
    return 0;			/* never reach */
  }
}

/* make singleton set of kind sp */

PRIVATE disj_set_t make_set(int id, an_sp_bag_kind sp, char * file, int line)
{
  disj_set_t s = alloc_disj_set();
  s->id = id;
  s->rank = 0;
  s->SP = sp;
  s->link = s;
#if defined RD_DEBUG
  s->children = NULL;
#endif /* RD_DEBUG */

  s->file = file;
  s->line = line;
  return s;
}

#if defined RD_DEBUG
PRIVATE void child_link_space_gc()
{
  fprintf(stderr, "child link space exhausted\n");
  fatal_error_exit();
}

PRIVATE children_set_t alloc_new_child_link()
{
  children_set_t p;
  children_set_t l;
  if (rd_globals_inited == 0) init_rd_globals();
  p = child_link_space_p;
  l = child_link_space_lim;
  if (p < l) {
    child_link_space_p = p + 1;
    return p;
  } else {
    child_link_space_gc();
  }
}

PRIVATE void add_child(disj_set_t parent, disj_set_t child)
{
  children_set_t l = alloc_new_child_link();
  
  l->a_child = child;
  l->other_children = parent->children;
  parent->children = l;
}

PRIVATE void remove_child(disj_set_t parent, disj_set_t child)
{
  children_set_t * l;
  
  for (l = &parent->children; *l != NULL; l = &(*l)->other_children) {
    if ((*l)->a_child == child) {
      (*l)->a_child == NULL;
      *l = (*l)->other_children;
      break;
    }
  }
}
#endif /* RD_DEBUG */

PRIVATE disj_set_t link(disj_set_t a, disj_set_t b, an_sp_bag_kind sp)
{
  if (a->rank < b->rank) {
    a->link = b; 
#if defined RD_DEBUG
    add_child(b, a);
#endif /* RD_DEBUG */
    b->SP = sp;
    return b;
  } else if (a->rank > b->rank) {
    b->link = a;
#if defined RD_DEBUG
    add_child(a, b);
#endif /* RD_DEBUG */
    a->SP = sp;
    return a;
  } else {			/* a->rank == b->rank */
    a->link = b;
#if defined RD_DEBUG
    add_child(b, a);
#endif /* RD_DEBUG */
    b->rank++;
    b->SP = sp;
    return b;
  }
}

PRIVATE inline disj_set_t find_set(disj_set_t s)
{
  if (RD_IS_EMPTY_SET(s)) return RD_EMPTY_SET;
  else {
    disj_set_t l = s->link;
    if (l == s) return s;
    else {
      disj_set_t r = find_set(l);
#if defined RD_DEBUG
      remove_child(s->link, s);
#endif /* RD_DEBUG */
      s->link = r;
#if defined RD_DEBUG
      add_child(r, s);
#endif /* RD_DEBUG */
      return r;
    }
  }
}

PRIVATE disj_set_t set_union(disj_set_t a, disj_set_t b, an_sp_bag_kind sp)
{
  if (RD_IS_EMPTY_SET(a)) {
      b->SP = sp;
      return b;
  } else if (RD_IS_EMPTY_SET(b)) {
      a->SP = sp;
      return a;
  } else {
    disj_set_t r = link(find_set(a), find_set(b), sp);
    return r;
  }
}

/*
 * Invoked when the application forks a statement
 */

GLOBAL void __rd_raw_fork_begin(
#if MEM_TRACE_PUT_ID
				source_id_t source_id, 
#endif
				char * file, int line)
{
  mtc_block_t b;
  int id;
  if (rd_dont_race_detect) return;
  if (rd_globals_inited == 0) init_rd_globals();
  b = active_block_stack_p + 1;
  id = unique_fork_id + 1;
  b->kind = mbk_fork;
  b->id_set = make_set(id, sbk_S, file, line); /* singleton set */
  b->S = b->id_set;		/* initially S-bag is singleton */
  b->P = RD_EMPTY_SET;		/* and P-bag is empty */

#if MEM_TRACE_PUT_ID
  b->source_id = source_id;
#endif
  active_block_stack_p++;
  unique_fork_id++;
}

/*
 * Invoked when the application exits a forked statement
 */

GLOBAL void __rd_raw_fork_end()
{
  mtc_block_t cb;		/* current block */
  mtc_block_t pb;		/* parent block */
  if (rd_dont_race_detect) return;
  if (rd_globals_inited == 0) init_rd_globals();
  cb = active_block_stack_p; 
  pb = active_block_stack_p - 1;
  pb->P = set_union(pb->P, cb->S, sbk_P);
  pb->P = set_union(pb->P, cb->P, sbk_P);
  cb->S = RD_EMPTY_SET;
  cb->P = RD_EMPTY_SET;

  active_block_stack_p--;
}

/*
 * Invoked when the application begins a sync block
 */

GLOBAL void __rd_sync_begin()
{
  mtc_block_t cb;		/* current block */
  mtc_block_t nb;		/* new block */
  if (rd_dont_race_detect) return;
  if (rd_globals_inited == 0) init_rd_globals();
  cb = active_block_stack_p; /* current block */
  nb = active_block_stack_p + 1; /* new block */
  nb->kind = mbk_sync;
  nb->id_set = cb->id_set;
  nb->S = RD_EMPTY_SET;
  nb->P = RD_EMPTY_SET;
  active_block_stack_p++;
}

/*
 * Invoked when the application exits a sync block
 */

GLOBAL void __rd_sync_end(char * file, int line)
{
  mtc_block_t cb;		/* current block */
  mtc_block_t pb;		/* parent block */
  if (rd_dont_race_detect) return;
  if (rd_globals_inited == 0) init_rd_globals();
  cb = active_block_stack_p;
  pb = active_block_stack_p - 1;
  if (!RD_IS_EMPTY_SET(cb->P)) {
    fprintf(stderr, 
	    "some threads have not been synched when sync block exits (%s:%d)\n", 
	    file, line);
    fatal_error_exit();
  }
  pb->S = set_union(pb->S, cb->S, sbk_S);
  cb->S = RD_EMPTY_SET;

  active_block_stack_p--;
}

/*
 * Invoked when the application performs a synchronization (---)
 */

GLOBAL void __rd_do_sync()
{
  mtc_block_t b;
  if (rd_dont_race_detect) return;
  if (rd_globals_inited == 0) init_rd_globals();
  b = active_block_stack_p;
  b->S = set_union(b->S, b->P, sbk_S);
  b->P = RD_EMPTY_SET;
}

/*
 * keep track of who read/wrote where
 */

/* Ensure ptable[p] is non-nil */
PRIVATE void ensure_primary_idx(long p)
{
  if (TR_ptable[p] == 0) {
    int i;
    TR_secondary_idx_t si = (TR_secondary_idx_t)
      mtc_general_alloc(sizeof(struct TR_secondary_idx));
    TR_ptable[p] = si;
    for (i = 0; i < TR_SECONDARY_IDX_MASK + 1; i++) {
      si->idx[i] = 0;
    }
  }
}

/* ensure ptable[p]->idx[q] is non-nil */
PRIVATE void ensure_secondary_idx(long p, long q)
{
  ensure_primary_idx(p);
  if (TR_ptable[p]->idx[q] == 0) {
    void * pp;
    void * ph = (char *)mtc_general_alloc(TR_trace_page_size);
    void * lim = ph + TR_trace_page_size;
    TR_ptable[p]->idx[q] = ph;
    for (pp = ph + TR_trace_page_overhead; 
	 pp < lim; 
	 pp += TR_trace_record_size) {
      an_access_kind k;
      TR_trace_record_t r = (TR_trace_record_t) pp;
      for (k = ak_read; k < ak_last; k++) {
	r->rec[k].th = 0;
	r->rec[k].file = 0;
	r->rec[k].line = 0;
      }
    }
  }
}

/* locate trace record for the addres A */
PRIVATE void * TR_table_walk(void * a)
{
  long p = TR_PRIMARY_IDX(a);
  long q = TR_SECONDARY_IDX(a);
  long k = TR_TLB_KEY(a);
  void * r;
  ensure_secondary_idx(p, q);
  r = TR_ptable[p]->idx[q];
  TR_TLB[k].tag = a;
  TR_TLB[k].val = r;
  return r;
}

/* generic routine that either reads or writes either reader or writer */
typedef enum a_get_or_set
{
  gs_get,
  gs_set
} a_get_or_set;

/* return the trace record for address A */
PRIVATE inline TR_trace_record_t get_trace_record(void * a)
{
  void * a_prime;		/* aligned address */
  long k;			/* TLB key */
  TR_tlb_entry_t ent;		/* address of the TLB entry */
  void * rec;			/* address of the record for this page */
  int offs;			/* offset within a page */
  int line_idx;			/* offs / line_size */
  TR_trace_record_t tr;		/* base of the trace record */

  /* lookup TLB for this page */
  a_prime = (void *)(((long)a) & ~(TR_TRACE_GRANULARITY - 1));
  k = TR_TLB_KEY(a_prime);
  ent = TR_TLB + k;
  if (ent->tag == a_prime) {
    rec = ent->val;		/* TLB hit */
  } else {
    rec = TR_table_walk(a_prime); /* full page table walk */
  }

  /* now look for the read address of the trace record */
  offs = TR_TLB_OFFS(a);
  line_idx = (offs >> TR_LOG_TRACE_GRANULARITY);
  tr = (TR_trace_record_t)(rec + (line_idx << TR_log_trace_record_size));

  return tr;
}

/* 1 if two race records R and S are the same */
PRIVATE int race_record_equal(race_record_t r, race_record_t s)
{
  if (r->si[0].ac.id != s->si[0].ac.id) return 0;
  if (r->si[1].ac.id != s->si[1].ac.id) return 0;
  if (r->si[0].th.id != s->si[0].th.id) return 0;
  if (r->si[1].th.id != s->si[1].th.id) return 0;

  if (r->si[0].ac.line != s->si[0].ac.line) return 0;
  if (r->si[1].ac.line != s->si[1].ac.line) return 0;
  if (r->si[0].th.line != s->si[0].th.line) return 0;
  if (r->si[1].th.line != s->si[1].th.line) return 0;

  if (r->si[0].ac.file == s->si[0].ac.file &&
      r->si[1].ac.file == s->si[1].ac.file &&
      r->si[0].th.file == s->si[0].th.file && 
      r->si[1].th.file == s->si[1].th.file) return 1;
  
  if (r->si[0].ac.file != s->si[0].ac.file &&
      strcmp(r->si[0].ac.file, s->si[0].ac.file) != 0) return 0;
  if (r->si[1].ac.file != s->si[1].ac.file &&
      strcmp(r->si[1].ac.file, s->si[1].ac.file) != 0) return 0;
  if (r->si[0].th.file != s->si[0].th.file &&
      strcmp(r->si[0].th.file, s->si[0].th.file) != 0) return 0;
  if (r->si[1].th.file != s->si[1].th.file &&
      strcmp(r->si[1].th.file, s->si[1].th.file) != 0) return 0;

  return 1;
}

/* hash for a race record */
PRIVATE int race_record_hash(race_record_t r)
{
  int a = ((r->si[0].th.id) >> RACE_RECORD_HASH_IGNORE) & RACE_RECORD_HASH_MASK;
  int b = ((r->si[0].ac.id) >> RACE_RECORD_HASH_IGNORE) & RACE_RECORD_HASH_MASK;
  int c = ((r->si[1].ac.id) >> RACE_RECORD_HASH_IGNORE) & RACE_RECORD_HASH_MASK;
  int d = ((r->si[1].th.id) >> RACE_RECORD_HASH_IGNORE) & RACE_RECORD_HASH_MASK;
  return (a + b + c + d) & RACE_RECORD_HASH_MASK;
}

PRIVATE int record_race(race_record_t r)
{
  if (races_found == 0) {
    int s = sizeof(race_record_t) * RACE_RECORD_HASH_SIZE;
    races_found = (race_record_t *)mtc_general_alloc(s);
    bzero((void *)races_found, s);
  }
  {
    /* look if the same record is already in the table */
    race_record_t * h = races_found;
    int k = race_record_hash(r);
    race_record_t p, q;
    race_record_t new_record;
    q = 0;
    for (p = h[k]; p; p = p->next) {
      if (race_record_equal(p, r)) {
	return 0; /* found. do not record R */
      }
      q = p;
    }
    /* the same record is not found in the hash. create and add a new record */
    new_record = (race_record_t)mtc_general_alloc(sizeof(struct race_record));
    bcopy((void *)r, (void *)new_record, sizeof(struct race_record));

    new_record->next = 0;
    if (q == 0) {
      h[k] = new_record;
    } else {
      q->next = new_record;
    }
    return 1;
  }
}


/* a kind of dependence */
typedef enum a_dependence_kind {
  dk_flow_dependence,
  dk_anti_dependence,
  dk_output_dependence
} a_dependence_kind;

/* a race is detected between the current access and the previous access
   designated by P.

   p->th->source_id : source id of the thread that performed the previous accsss
   p->th->file :      where it was created (file name)
   p->th->line :      where it was created (line number)
   p->source_id :     source id of the previous accss
   p->file :          where did the access occur (file name)
   p->line :          where did the access occur (line number)

   th->source_id :    source id of the current thread
   th->file :         where the current thread was created (file name)
   th->line :         where the current thread was created (line number)
   source_id :        source id of the current access
   file :             where did the current access occur (file name)
   line :             where did the current access occur (line number)
*/
PRIVATE void race_detected(a_dependence_kind dk, 
			   void * addr,
			   TR_access_record_t p	/* previous access */,
			   disj_set_t th, 
#if MEM_TRACE_PUT_ID
			   /* source identifier where this access occurred */
			   source_id_t source_id, 
#endif /* MEM_TRACE_PUT_ID */
			   char * file, int line)
{
  struct race_record rr[1];
#if MEM_TRACE_PUT_ID
  rr->si[0].th.id = p->th->source_id;
#endif /* MEM_TRACE_PUT_ID */
  rr->si[0].th.file = p->th->file;
  rr->si[0].th.line = p->th->line;
#if MEM_TRACE_PUT_ID
  rr->si[0].ac.id = p->source_id;
#endif /* MEM_TRACE_PUT_ID */
  rr->si[0].ac.file = p->file;
  rr->si[0].ac.line = p->line;
#if MEM_TRACE_PUT_ID
  rr->si[1].th.id = th->source_id;
#endif /* MEM_TRACE_PUT_ID */
  rr->si[1].th.file = th->file;
  rr->si[1].th.line = th->line;
#if MEM_TRACE_PUT_ID
  rr->si[1].ac.id = source_id;
#endif /* MEM_TRACE_PUT_ID */
  rr->si[1].ac.file = file;
  rr->si[1].ac.line = line;

  if (record_race(rr)) {
#if 1
    fprintf(stdout, "mtcamp_race_detected: %s:%d %s:%d %s:%d %s:%d \n",
	    rr->si[0].th.file, rr->si[0].th.line,
	    rr->si[0].ac.file, rr->si[0].ac.line,
	    rr->si[1].th.file, rr->si[1].th.line,
	    rr->si[1].ac.file, rr->si[1].ac.line);

#else  /* 1 */
    char * s = 0;
    disj_set_t curr_th = active_block_stack_p->id_set;
    disj_set_t prev_th = p->th;
    switch (dk) {
    case dk_flow_dependence: 
      s = "flow dependence";
      break;
    case dk_anti_dependence:
      s = "anti dependence";
      break;
    case dk_output_dependence:
      s = "output dependence";
      break;
    default:
      fprintf(stderr, "unknown dependence kind = %d\n", dk);
      fatal_error_exit();
    }

    fprintf(stderr, "%s detected at address %p\n", s, addr);

  /* report where the previous accessor is created and where it accesses
     the location */
    fprintf(stderr, "thread %d ", prev_th->id);
    if (prev_th->id != INITIAL_THREAD_ID) {
      fprintf(stderr, "was created at (%s:%d)", prev_th->file, prev_th->line);
    } else {
      fprintf(stderr, "was created as the main thread");
    }
    fprintf(stderr, " and accessed %p at (%s:%d)\n", addr, p->file, p->line);

  /* report where the current accessor is created and where it accesses
     the location */
    fprintf(stderr, "thread %d ", curr_th->id);
    if (curr_th->id != INITIAL_THREAD_ID) {
      fprintf(stderr, "was created at (%s:%d)", curr_th->file, curr_th->line);
    } else {
      fprintf(stderr, "was created as the main thread");
    }
    fprintf(stderr, " and accessed %p at (%s:%d)\n", addr, file, line);
#endif /* 1 */

#if STOP_AFTER_FIRST_RACE
    exit(1);
#endif
  }
}

/* 
 * Invoked when application reads an address
 */

GLOBAL void __mem_trace_read(void * addr, int size, 
#if MEM_TRACE_PUT_ID
			     source_id_t source_id,
#endif /* MEM_TRACE_PUT_ID */
			     char * file, int line)
{
  void * a; void * limit = addr + size;
  if (rd_dont_race_detect) return;
  if (rd_globals_inited == 0) init_rd_globals();

  for (a = addr; a < limit; a += TR_TRACE_GRANULARITY) {
    TR_trace_record_t tr = get_trace_record(a);
    TR_access_record_t pr, pw;
    disj_set_t ws, rs;
      
      /* obtain previous writer record */
    pw = tr->rec + ak_write;
    ws = find_set(pw->th);
    if ((!RD_IS_EMPTY_SET(ws)) && RD_IS_PBAG(ws)) {
      race_detected(dk_flow_dependence, addr,
		    pw, active_block_stack_p->id_set, 
#if MEM_TRACE_PUT_ID
		    source_id, 
#endif /* MEM_TRACE_PUT_ID */
		    file, line);
    } 
    
    /* obtain reader record */
    pr = tr->rec + ak_read;
    rs = find_set(pr->th);
    if (RD_IS_EMPTY_SET(rs) || RD_IS_SBAG(rs)) {
      /* set reader record */
      pr->th = active_block_stack_p->id_set;
#if MEM_TRACE_PUT_ID
      pr->source_id = source_id;
#endif /* MEM_TRACE_PUT_ID */
      pr->file = file;
      pr->line = line;
    }
  }
}

/* 
 * Invoked when application writes an address
 */

GLOBAL void __mem_trace_write(void * addr, int size, 
#if MEM_TRACE_PUT_ID
			      source_id_t source_id,
#endif
			      char * file, int line)
{
  void * a; void * limit = addr + size;
  if (rd_dont_race_detect) return;
  if (rd_globals_inited == 0) init_rd_globals();

  for (a = addr; a < limit; a += TR_TRACE_GRANULARITY) {
    TR_trace_record_t tr = get_trace_record(a);
    TR_access_record_t pr, pw;
    disj_set_t ws, rs;
    
    /* obtain writer record */
    pw = tr->rec + ak_write;
    ws = find_set(pw->th);
    if ((!RD_IS_EMPTY_SET(ws)) && RD_IS_PBAG(ws)) {
      race_detected(dk_output_dependence, addr,
		    pw, active_block_stack_p->id_set, 
#if MEM_TRACE_PUT_ID
		    source_id, 
#endif /* MEM_TRACE_PUT_ID */
		    file, line);
    } 
    
    /* obtain reader record */
    pr = tr->rec + ak_read;
    rs = find_set(pr->th);
    if ((!RD_IS_EMPTY_SET(rs)) && RD_IS_PBAG(rs)) {
      race_detected(dk_anti_dependence, addr,
		    pr, active_block_stack_p->id_set, 
#if MEM_TRACE_PUT_ID
		    source_id, 
#endif /* MEM_TRACE_PUT_ID */
		    file, line);
    } 
    
    /* set writer */
    pw->th = active_block_stack_p->id_set;
#if MEM_TRACE_PUT_ID
    pw->source_id = source_id;
#endif /* MEM_TRACE_PUT_ID */
    pw->file = file;
    pw->line = line;
  }
}

/* 
 * Invoked when application reads and writes an address
 */

GLOBAL void __mem_trace_read_write(void * addr, int size, 
#if MEM_TRACE_PUT_ID
				   source_id_t source_id,
#endif
				   char * file, int line)
{
  __mem_trace_read(addr, size, 
#if MEM_TRACE_PUT_ID
		   source_id,
#endif
		   file, line);
  __mem_trace_write(addr, size, 
#if MEM_TRACE_PUT_ID
		    source_id,
#endif
		    file, line);
}

/*
 * 
 */

GLOBAL void __rd_mem_alloced(void * addr, int size, char * file, int line)
{
  void * a; void * limit = addr + size;
  if (rd_dont_race_detect) return;
  if (rd_globals_inited == 0) init_rd_globals();

  for (a = addr; a < limit; a += TR_TRACE_GRANULARITY) {
    TR_trace_record_t tr = get_trace_record(a);
    TR_access_record_t pr, pw;
    
    /* clear writer record */
    pw = tr->rec + ak_write;
    pw->th = RD_EMPTY_SET;
    pw->file = file;
    pw->line = line;

    /* clear reader record */
    pr = tr->rec + ak_read;
    pr->th = RD_EMPTY_SET;
    pr->file = file;
    pr->line = line;

  }
}

GLOBAL void __rd_mem_freed(void * addr, char * file, int line)
{
  
}


